require 'configatron'
require 'configatron_plus/errors'
require 'pathname'
module ConfigatronPlus

  DEFAULT_SOURCES = [:current, :env]

  def self.config
    configatron.configatron_plus
  end

  configatron.configatron_plus do |c|
    c.set_default(:sources,           DEFAULT_SOURCES.dup)
    c.set_default(:rc_name,           "config.rb")
    c.set_default(:env_prefix,        "configatron_")
    c.set_default(:source_files,      {:current => configatron.configatron_plus.rc_name})
    c.set_default(:fail_when_missing, false)
  end

  def self.fail_when_missing
    configatron.configatron_plus.fail_when_missing
  end

  def self.fail_when_missing=(val)
    configatron.configatron_plus.fail_when_missing = val
  end

  # Set the sources to use for configatron
  # @param new_val [Array<String,Symbol>] new array of sources.
  def self.sources=(new_val)
    @@errors = nil
    raise ArgumentError.new "value must be an Array." unless new_val.is_a?(Array)
    unless self.valid_list?(new_val)
      raise ArgumentError.new "Array members must be a symbol, string, or Pathname: #{self.errors}"
    end
     self.config.sources = new_val
  end

  # Get the current list of sources
  # @return [Array<String,Symbol>] of sources to look for configurations in
  def self.sources
    self.config.sources
  end

  # @return [Array] of symbols.

  def self.errors
    @@errors ||= []
  end

  #set @@sources = DEFAULT_SOURCES
  def self.reset_sources
    self.sources = DEFAULT_SOURCES.dup
  end

  #check each element of the list to see if it's a correct symbol or can be used a a filename.
  # @return [Boolean]
  def self.valid_list?(val)
    val.each do |i|
      errors << "Invalid entry: #{i}" unless self.can_use_as_filename?(i)
      if i.is_a?(Symbol) && ! self.is_symbol_defined?(i)
        errors << "Invalid symbol: #{i.to_s}"
      end
    end
    errors.empty?
  end

  def self.is_symbol_defined?(sym)
    return true if sym == :env
    self.config.source_files.keys.include?(sym)
  end

  # @param opts [Hash] the options for getting configuration
  # @option opts [Boolean] :fail_on_missing (false) Raise an exception if a file in the @sources list is not present
  def self.fetch_sources
    self.sources.each do |s|
      if s.is_a?(Symbol)
        get_from_special(s)
      else
        get_from_file(s.is_a?(Pathname) ? s : Pathname(s))
      end
    end
  end

  # Checks to see if we can use this as a filename
  # @param val [Object] Object to check
  # @return [Boolean]]
  def self.can_use_as_filename?(val)
    return true if val.is_a?(String)
    return true if val.is_a?(Pathname)
    return true if val.is_a?(Symbol)

    #if no other
    false
  end

  # looks upo the file for a given symbol and calls get_from_file with it
  # @param s [Symbol] use to look up and load a file
  def self.get_from_special(s)
    raise ArgumentError unless s.is_a?(Symbol)
    return get_from_env if s == :env
    file = config.source_files.fetch(s,nil)
    raise ConfigatronPlus::ConfigError.new("Setting does not exist for: #{s}") if file.nil?
    file = Pathname(file)
    self.get_from_file(file)
  end

  # checks to see if file exists and loads it
  # @param file [Pathname] file to attempt to load
  def self.get_from_file(file)
    raise ArgumentError.new("Not Pathname: #{file}") unless file.is_a?(Pathname)
    if file.expand_path.file?
      c = File.read(file.expand_path)
      eval(c)
    elsif self.fail_when_missing
      raise ConfigatronPlus::ConfigError.new("Config file does not exist: #{file.to_s}")
    end
  end

  #scan ENV for any keys that start with alfred_
  #check to see that they are valid names
  #add them to configatron
  def self.get_from_env
    ENV.keys.find_all {|e| self.use_this_env?(e)}.each do |key|
      setting = key[config.env_prefix.length .. -1]
      setting.split('_').each do |e|
        raise ConfigatronPlus::ConfigError.new("Env variable '#{key}' contains invalid characters") unless self.is_valid_key_name?(e)
      end
      eval("configatron." + setting.gsub('_','.') + " = \"#{ENV[key]}\"")
    end
  end

  def self.use_this_env?(env)
    env.length > config.env_prefix.length && env.start_with?(config.env_prefix)
  end

  # Checks a string to make sure it is a valid identifier so we can eval it
  # @param e [String] string to check for validity
  # @return [Boolean] is valid?
  def self.is_valid_key_name?(e)
    /^[a-z][a-z0-9]*$/.match(e.downcase)
  end
end